/*
* $Id$
*
*****************************************************************************
*
* MODULE:   	r.contour
* AUTHOR(S):	Terry Baker US Army Construction Engineering Research Lab
*   	    	Justin Hickey - Thailand - jhickey@hpcc.nectec.or.th
* PURPOSE:  	To generate contours for a raster map. Needs more...
*
*   	    	To keep track of which points have already been used as
*   	    	contour points a "hits" hash table is used. The hash table
*   	    	is simply an array of pointers that will point to one or
*   	    	more hash nodes depending on whether there are any collisions.
*   	    	A problem with the hash table in this implementation is that it
*   	    	needs to be initialized for each contour level. In an attempt
*   	    	to save time in memory allocation and freeing, a pool of hash
*   	    	nodes is used.
*
*   	    	The idea is to have a chunk of memory allocated as a pool of
*   	    	hash nodes. Then, a pointer (freePtr) can be used to point to
*   	    	the next free node. As nodes are used, the pointer is
*   	    	incremented. When the pointer reaches the last node, a new
*   	    	chunk of memory is allocated. The pool is not reallocated
*   	    	since many pointers would need to be updated for all previous
*   	    	nodes. To link the previous nodes with the new nodes, the last
*   	    	node is used as a special node which points to the start of the
*   	    	new chunk of memory. This special node is never available as a
*   	    	pool node.
*
*   	    	This method simplifies initialization in that only the hash
*   	    	table pointers need to be initialized to NULL for each contour
*   	    	level, and the free pointer is reset to the beginning of the
*   	    	initial chunk of memory. The actual nodes do not need to be
*   	    	cleared since they will be re-initialized when they are used
*   	    	again.
*
*   	    	Although this method is rather complex and a bit messy, it
*   	    	should save quite a bit of time in memory maintenance.
* COPYRIGHT:    (C) 2001 by the GRASS Development Team
*
*               This program is free software under the GNU General Public
*   	    	License (>=v2). Read the file COPYING that comes with GRASS
*   	    	for details.
*
*****************************************************************************/

/*============================= Include Files ==============================*/

/* System include files */
#include <stdio.h>
#include <string.h>
#include <math.h>

/* Grass and local include files */
#include "gis.h"
#include "Vect.h"
#include "dig_atts.h"
#include "local_proto.h"

/*======================= Internal Constants/Defines =======================*/

/* See comments in AllocateHashTable() and GetHashIndex() */
static const int    HASH_SIZE = 512;
static const double HASH_CONST = 0.6180339887; /* (sqrt(5.0) - 1.0) / 2.0 */

/*========================== Internal Typedefs =============================*/

typedef enum Edge
{   
    TOP_EDGE,
    RIGHT_EDGE, 
    BOTTOM_EDGE,
    LEFT_EDGE
} Edge;

typedef enum NodeType
{   
    NO_TYPE,
    CONTOUR_WINDOW,
    CONTOUR_POINT,
    END_MEM_CHUNK
} NodeType;

typedef enum EndPoint
{   
    NO_END_PNT,
    START_LINE,
    END_LINE
} EndPoint;

typedef struct ContourCell
{
    DCELL   z[4];
    int     row;
    int     col;
    Edge    edge;
} ContourCell;

typedef struct HashNode
{
    double  	    x;
    double  	    y;
    EndPoint	    endPnt;
    NodeType   	    type;
    struct HashNode *next;
} HashNode;

/*====================== Static Variable Declaration =======================*/

/* Hash table variables */
static HashNode     **hits;
static char **hit;

/* Hash node pointers and variables for maintaining the pool of nodes */
static HashNode     *headPtr = NULL;
static HashNode     *freePtr = NULL;

/*============================== Prototypes ================================*/

static void AllocateHashTable(void);
static void AllocateNodePool(void);
static void InitHashTable (void);
static int GetHashIndex (double, double);
static void AddHashNode (double, double, NodeType);
static int HaveHit (double, double, NodeType);

static int GetNewCell (ContourCell *, int, int, DCELL **, double,
    struct Cell_head, struct line_pnts *, int, int);
static int ChkMapEdge (ContourCell *, int, int, DCELL **, double,
    struct Cell_head, struct line_pnts *, int *, int, int);
static int CheckPos (ContourCell *, int, int);
static void ResetCell (ContourCell *, DCELL **);
static void NewEdge (ContourCell *);
static void SetEdge (ContourCell *, int);
static int FindCrossing (ContourCell *, double, struct Cell_head,
    struct line_pnts *, int);
static void AddPoint (int, ContourCell *, double, struct Cell_head,
    struct line_pnts *);
static int CheckEdge (DCELL, DCELL, double);

/*======================= Internal Static Functions ========================*/

/****************************************************************************
* void AllocateHashTable (void)
*
* PURPOSE: 	To allocate the hash table for the hits array. Note that the
*   	    	size of the hash table is currently set to 512. This value is
*   	    	arbitrary and was chosen mainly since it was a power of 2 that
*   	    	could quickly be initialized all to null, and even with 5000
*   	    	points in contour lines gives an average of 10 collisions in
*   	    	the hash table.
* INPUT VARS:	none
* RETURN VAL:	none
*****************************************************************************/
static void AllocateHashTable (void)
{
    /* Allocate the memory for the hash table */
    hits = (HashNode **) G_malloc(HASH_SIZE * sizeof(HashNode *));
    
    return;
}

/****************************************************************************
* void AllocateNodePool (void)
*
* PURPOSE: 	To allocate a chunk of memory to use as part of the node pool.
*   	    	Each call to this function will allocate a chunk of memory and
*   	    	the end of the previous memory will be linked to the newly
*   	    	allocated memory if necessary.
* INPUT VARS:	none
* RETURN VAL:	none
*****************************************************************************/
static void AllocateNodePool (void)
{
    static HashNode *endMemory;
    HashNode	    *tempPtr;
    HashNode	    *memChunk;
    int     	    i;
    
    /* Allocate a chunk of memory */
    memChunk = (HashNode *) G_malloc(HASH_SIZE * sizeof(HashNode));
    
    /* Check if this is the first allocation - no previous end node */
    if (headPtr == NULL)
    {
    	headPtr = memChunk;
    }
    else
    {
    	/* Link the end of the previous memory chunk to the new chunk */
    	endMemory->next = memChunk;
    }
    
    /* Since the "type of node" is used to mark the end of this chunk of */
    /* memory, the type field must be cleared first to ensure that an actual */
    /* node is not interpreted as an end node. This can happen if by chance, */
    /* the value of the type memory location of a node just happens to be */
    /* equal to END_MEM_CHUNK when the memory is allocated */
    tempPtr = memChunk;
    
    for (i = 0; i < HASH_SIZE; i++)
    {
    	tempPtr->type = NO_TYPE;
	tempPtr++;
    }
    
    /* Set up the last node of the new memory chunk to be an end marker */
    /* Note that -99.0 is an arbitrary dummy value */
    endMemory = memChunk + (HASH_SIZE - 1);
    endMemory->x = -99.0;
    endMemory->y = -99.0;
    endMemory->endPnt = NO_END_PNT;
    endMemory->type = END_MEM_CHUNK;
    endMemory->next = NULL;
    
    /* Set the free pointer to the start of the new memory */
    freePtr = memChunk;
    
    return;
}

/****************************************************************************
* void InitHashTable (void)
*
* PURPOSE: 	To 
* INPUT VARS:	none
* RETURN VAL:	none
*****************************************************************************/
static void InitHashTable (void)
{
    char    errMsg[512];    /* error message */
    int     i;
    
    /* Check if memory has been allocated */
    if (hits == NULL || headPtr == NULL)
    {
    	strcpy(errMsg, "Memory not allocated when calling InitHashTable. ");
	strcat(errMsg, "Please advise GRASS developers of this error.");
    	G_fatal_error("%s", errMsg);
    }
    
    /* Set all the hash table pointers to NULL */
    for (i = 0; i < HASH_SIZE; i++)
    {
    	hits[i] = NULL;
    }
    
    /* Set the free pointer to the beginning of the memory chain */
    freePtr = headPtr;
    
    return;
}

/****************************************************************************
* int GetHashIndex(double x, double y)
*
* PURPOSE: 	To determine the hash index given the x and y coordinates in
*   	    	raster coordinates, not UTM coordinates. This function uses
*   	    	the multiplication method for a hash function. That is,
*
*   	    	    h(k) = floor(m * (k * A - floor(k * A)))
*
*   	    	where
*   	    	    h(k)    =>	result of hash function
*   	    	    m	    =>	size of hash table
*   	    	    k	    =>	key of hash index
*   	    	    A	    =>	constant
*
*   	    	The value of m is usually a power of 2 to benefit from the
*   	    	binary nature of computers. The value of A is
*
*   	    	    (sqrt(5) - 1) / 2
*
*   	    	This value of A was recommended by Donald Knuth in "Sorting
*   	    	and Searching, vol 3 of The Art of Computer Programming, 1973"
*   	    	Note that (k * A - floor(k * A)) is simply the fractional part
*   	    	of k * A. This hash function comes from "Introduction to
*   	    	Algorithms, by Cormen, Leiserson, and Rivest, 1990". For the 
*   	    	purposes of the hit hash table,
*
*   	    	    k = x + y
*   	    	    m = 512
*
*   	    	The choice of m as 512 is arbitrary. See the
*   	    	AllocateHashTable() function for a brief reason for this value.
* INPUT VARS:	x
*   	    	y
* RETURN VAL:	index into hash table
*****************************************************************************/
static int GetHashIndex (double x, double y)
{
    double  keyVal;
    int     returnVal;
    
    /* Make sure the values are positive */
    x = fabs(x);
    y = fabs(y);
    
    keyVal = (x + y) * HASH_CONST;
    returnVal = (int) ((double) HASH_SIZE * (keyVal - (int) keyVal));
    
    return (returnVal);
}

/****************************************************************************
* void AddHashNode (double x, double y, NodeType type)
*
* PURPOSE: 	To 
* INPUT VARS:	none
* RETURN VAL:	none
*****************************************************************************/
static void AddHashNode (double x, double y, NodeType type)
{
    HashNode	*tempPtr;
    char    	errMsg[512];	/* error message */
    int     	hashIndex;
    
    /* Check if the hash table has been initialized */
    if (freePtr == NULL)
    {
    	strcpy(errMsg, "Hash table not initialized when trying to add node. ");
	strcat(errMsg, "Please advise GRASS developers of this error.");
    	G_fatal_error("%s", errMsg);
    }
    
    /* Check if we need more memory */
    if (freePtr->type == END_MEM_CHUNK)
    {
    	if (freePtr->next == NULL)
	{
	    AllocateNodePool();
	}
	else
	{
	    freePtr = freePtr->next;
	}
    }
    
    /* Initialize the new node */
    freePtr->x = x;
    freePtr->y = y;
    freePtr->endPnt = NO_END_PNT;
    freePtr->type = type;
    freePtr->next = NULL;
    
    /* Get the hash index */
    hashIndex = GetHashIndex(x, y);
    tempPtr = hits[hashIndex];
    
    /* Find where to place the new node */
    if (tempPtr != NULL)
    {
    	while (tempPtr->next != NULL)
	{
	    tempPtr = tempPtr->next;
	}
	
	tempPtr->next = freePtr;
    }
    else
    {
    	hits[hashIndex] = freePtr;
    }
    
    /* Increment the free pointer */
    freePtr++;
    
    return;
}

/****************************************************************************
* void HaveHit (double x, double y, NodeType type)
*
* PURPOSE: 	To 
* INPUT VARS:	none
* RETURN VAL:	none
*****************************************************************************/
static int HaveHit (double x, double y, NodeType type)
{
    HashNode	*tempPtr;
    char    	errMsg[512];	/* error message */
    int     	hashIndex;
    
    /* Check if the hash table has been initialized */
    if (freePtr == NULL)
    {
    	strcpy(errMsg, "Hash table not initialized when checking hit. ");
	strcat(errMsg, "Please advise GRASS developers of this error.");
    	G_fatal_error("%s", errMsg);
    }
    
    /* Get the hash index */
    hashIndex = GetHashIndex(x, y);
/*fprintf(stderr, "hash index = %d\n", hashIndex);*/
    tempPtr = hits[hashIndex];
    
    /* Check for a possible hit */
    if (tempPtr != NULL)
    {
/*fprintf(stderr, "Have a possible hit\n");*/
	/* Check the current node */
	if (tempPtr->x == x && tempPtr->y == y && tempPtr->type == type)
	{
	    return TRUE;
	}
	
	/* Scan for any collisions */
    	while (tempPtr->next != NULL)
	{
/*fprintf(stderr, "Scanning the collisions\n");*/
	    tempPtr = tempPtr->next;
	    
	    if (tempPtr->x == x && tempPtr->y == y && tempPtr->type == type)
	    {
	    	return TRUE;
	    }
	}
    }
    
    return FALSE;
}

/****************************************************************************
* int GetNewCell(ContourCell *current, int numRow, int numCol, DCELL **z,
*   double level, struct Cell_head cellHead, struct line_pnts *points,
*   int startRow, int startCol)
*
* PURPOSE: 	if cell is in range, finds data values of corner points of
*   	    	current cell returns FALSE else returns TRUE
* INPUT VARS:	current
*   	    	numRow
*   	    	numCol
*   	    	z
*   	    	level
*   	    	cellHead
*   	    	points
*   	    	startRow
*   	    	startCol
* RETURN VAL:	boolean
*****************************************************************************/
static int GetNewCell(ContourCell *current, int numRow, int numCol, DCELL **z,
    double level, struct Cell_head cellHead, struct line_pnts *points,
    int startRow, int startCol)
{
    int     pntAdded = FALSE;
    
    /* Check if we are still inside the map region */
    if (CheckPos(current, numRow, numCol))
    {
    	ResetCell(current, z);
	
	return FALSE;
    }
    else
    {

fprintf(stderr,"Checking for map edge\n");
    	while (!ChkMapEdge(current, numRow, numCol, z, level, cellHead, points,
	    &pntAdded, startRow, startCol))
    	{
	    /* Do nothing. The while loop is needed in case the contour */
	    /* following the map edge changes direction. In this case, we */
	    /* need to call ChkMapEdge() again with the new direction which */
	    /* is indicated by a change in the current->edge value. This */
	    /* value is changed inside ChkMapEdge() */
	    ;
	}

fprintf(stderr,"Done map edge check\n");

	/* Check if we are at the start of the contour */
	/* might need to change for hash code */
	if (pntAdded && !(HaveHit((double) current->col, (double) current->row,
    	    CONTOUR_WINDOW)))
	{
	    return FALSE;
	}
    }
    
    return TRUE;
}

/****************************************************************************
* int ChkMapEdge(ContourCell *current, int numRow, int numCol, DCELL **z,
*   double level, struct Cell_head cellHead, struct line_pnts *points,
*   int *pntAdded, int startRow, int startCol)
*
* PURPOSE: 	if cell is in range, finds data values of corner points of
*   	    	current cell returns FALSE else returns TRUE
* INPUT VARS:	current
*   	    	numRow
*   	    	numCol
*   	    	z
*   	    	level
*   	    	cellHead
*   	    	points
*   	    	pntAdded
*   	    	startRow
*   	    	startCol
* RETURN VAL:	boolean
*****************************************************************************/
static int ChkMapEdge(ContourCell *current, int numRow, int numCol, DCELL **z,
    double level, struct Cell_head cellHead, struct line_pnts *points,
    int *pntAdded, int startRow, int startCol)
{
    double  x;
    double  y;
    int     nextRow;
    int     nextCol;
    
    /* Check if contour line should continue along the edge of the */
    /* map region, that is if z == level */
    switch (current->edge)
    {
	/* We've crossed the bottom border and edge is to the top */
	case TOP_EDGE:

fprintf(stderr,"top\n");
	    /* Check if the last cell was exactly equal to level. Note that */
	    /* the last cell position depends on whether we are following */
	    /* the contour in a clockwise, or counter clockwise direction so */
	    /* 2 checks are needed. First check clockwise */
	    if (z[current->row][current->col] == level)
	    {
fprintf(stderr, "clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row;
		nextCol = current->col - 1;
		
		/* Check if the cell to the left is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the bottom edge */
			current->col = current->col - 1;
		    	nextCol = current->col - 1;

			/* Check if we are at the map's left edge */
			if (current->col <= 0)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = numRow - 1;
			    x = 0;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the left */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->col = -1;
			    current->edge = RIGHT_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = BOTTOM_EDGE;
		    current->row = current->row - 1;
		    current->col = current->col - 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }
	    
	    /* Second check counter clockwise */
	    else if (current->col < numCol - 1 &&
	    	z[current->row][current->col + 1] == level)
	    {
fprintf(stderr, "counter clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row;
		nextCol = current->col + 2;
		
		/* Check if the cell to the right is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the bottom edge */
			current->col = current->col + 1;
		    	nextCol = current->col + 2;

			/* Check if we are at the map's right edge */
			if (current->col >= numCol - 2)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = numRow - 1;
			    x = numCol - 1;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the right */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = numRow - 1;
			    current->col = numCol - 1;
			    current->edge = LEFT_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = LEFT_EDGE;
		    current->row = current->row - 1;
		    current->col = current->col + 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }

	    break;

	/* We've crossed the left border and edge is to the right */
	case RIGHT_EDGE:

fprintf(stderr,"right\n");
	    /* Check if the last cell was exactly equal to level. Note that */
	    /* the last cell position depends on whether we are following */
	    /* the contour in a clockwise, or counter clockwise direction so */
	    /* 2 checks are needed. First check clockwise */
	    if (current->col < numCol - 1 &&
	    	z[current->row][current->col + 1] == level)
	    {
fprintf(stderr, "clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row - 1;
		nextCol = current->col + 1;
		
		/* Check if the cell above is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the left edge */
			current->row = current->row - 1;
		    	nextRow = current->row - 1;

			/* Check if we are at the map's top edge */
			if (current->row <= 0)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = 0;
			    x = 0;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the top */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = -1;
			    current->col = -1;
			    current->edge = BOTTOM_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = BOTTOM_EDGE;
		    current->row = current->row - 1;
		    current->col = current->col + 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }
	    
	    /* Second check counter clockwise */
	    else if (current->row < numRow - 1 && current->col < numCol - 1 &&
	    	z[current->row + 1][current->col + 1] == level)
	    {
fprintf(stderr, "counter clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row + 1;
		nextCol = current->col + 1;
		
		/* Check if the cell below is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the left edge */
			current->row = current->row + 1;
		    	nextRow = current->row + 1;

			/* Check if we are at the map's bottom edge */
			if (current->row >= numRow - 2)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = numRow - 1;
			    x = 0;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the bottom */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = numRow - 1;
			    current->col = -1;
			    current->edge = TOP_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = TOP_EDGE;
		    current->col = current->col + 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }

	    break;

	/* We've crossed the top border and edge is to the bottom */
	case BOTTOM_EDGE:

fprintf(stderr,"bottom\n");
	    /* Check if the last cell was exactly equal to level. Note that */
	    /* the last cell position depends on whether we are following */
	    /* the contour in a clockwise, or counter clockwise direction so */
	    /* 2 checks are needed. First check clockwise */
	    if (current->row < numRow - 1 && current->col < numCol - 1 &&
	    	z[current->row + 1][current->col + 1] == level)
	    {
fprintf(stderr, "clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row + 1;
		nextCol = current->col + 2;
		
		/* Check if the cell to the right is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
fprintf(stderr,"moving right\n");
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the top edge */
			current->col = current->col + 1;
		    	nextCol = current->col + 2;

			/* Check if we are at the map's right edge */
			if (current->col >= numCol - 2)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = 0;
			    x = numCol - 1;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the right */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = -1;
			    current->col = numCol - 1;
			    current->edge = LEFT_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = TOP_EDGE;
		    current->row = current->row + 1;
		    current->col = current->col + 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }
	    
	    /* Second check counter clockwise */
	    else if (current->row < numRow - 1 &&
	    	z[current->row + 1][current->col] == level)
	    {
fprintf(stderr, "counter clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row + 1;
		nextCol = current->col - 1;
		
		/* Check if the cell to the left is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
/*fprintf(stderr,"moving left\n");*/
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the top edge */
			current->col = current->col - 1;
		    	nextCol = current->col - 1;

			/* Check if we are at the map's left edge */
			if (current->col <= 0)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = 0;
			    x = 0;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the left */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = -1;
			    current->col = -1;
			    current->edge = RIGHT_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = TOP_EDGE;
		    current->row = current->row + 1;
		    current->col = current->col - 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }

	    break;

	/* We've crossed the right border and edge is to the left */
	case LEFT_EDGE:

fprintf(stderr,"left row = %d col = %d\n", current->row, current->col);
	    /* Check if the last cell was exactly equal to level. Note that */
	    /* the last cell position depends on whether we are following */
	    /* the contour in a clockwise, or counter clockwise direction so */
	    /* 2 checks are needed. First check clockwise */
	    if (current->row < numRow - 1 &&
	    	z[current->row + 1][current->col] == level)
	    {
fprintf(stderr, "clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row + 2;
		nextCol = current->col;
		
		/* Check if the cell below is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
/*fprintf(stderr,"below\n");*/
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the right edge */
			current->row = current->row + 1;
		    	nextRow = current->row + 2;

			/* Check if we are at the map's bottom edge */
			if (current->row >= numRow - 2)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = numRow - 1;
			    x = numCol - 1;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the bottom */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = numRow - 1;
			    current->col = numCol - 1;
			    current->edge = TOP_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = RIGHT_EDGE;
		    current->row = current->row + 1;
		    current->col = current->col - 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }
	    
	    /* Second check counter clockwise */
	    else if (z[current->row][current->col] == level)
	    {
fprintf(stderr, "counter clockwise\n");
		/* Get next row and column - needed to check for start point */
		nextRow = current->row - 1;
		nextCol = current->col;
		
		/* Check if the cell above is == level */
		if (!(HaveHit((double) nextCol, (double) nextRow,
		    CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		{
		    while (!(HaveHit((double) nextCol, (double) nextRow,
		    	CONTOUR_WINDOW)) && z[nextRow][nextCol] == level)
		    {
			/* Follow along the right edge */
			current->row = current->row - 1;
		    	nextRow = current->row - 1;

			/* Check if we are at the map's top edge */
			if (current->row <= 0)
			{
    	    	    	    /* We have a corner point so add it directly */
    			    y = 0;
			    x = numCol - 1;
			    
fprintf(stderr, "new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
			    AddHashNode(x, y, CONTOUR_POINT);
			    y = cellHead.north - (y + 0.5) * cellHead.ns_res;
    			    x = cellHead.west + (x + 0.5) * cellHead.ew_res;

    			    Vect_append_point(points, x, y);
			    
			    /* Set the edge to indicate we are off the top */
			    /* side of the map and return FALSE so we call */
			    /* ChkMapEdge() again */
			    current->row = -1;
			    current->edge = BOTTOM_EDGE;
			    return FALSE;
			}
		    }

		    /* Set the cell for the next point and add it */
		    current->edge = BOTTOM_EDGE;
		    current->row = current->row - 1;
		    current->col = current->col - 1;
    	    	    ResetCell(current, z);

    	    	    AddPoint(FALSE, current, level, cellHead, points);
    	    	    *pntAdded = TRUE;
		}
	    }

	    break;

	default:
	    G_fatal_error ("Illegal edge number while checking map edge");
	    break;
    }
    
    /* This should check for an xy hit not an index hit */
    if (nextRow == startRow && nextCol == startCol)
    {
	/* Connect the contour to the starting point */
	y = nextRow;
	x = nextCol;

fprintf(stderr, "Startpoint: new point row = %d, col = %d, x = %g, y = %g\n", current->row, current->col, x, y);
    	/*AddHashNode(x, y, CONTOUR_POINT);*/
	y = cellHead.north - (y + 0.5) * cellHead.ns_res;
	x = cellHead.west + (x + 0.5) * cellHead.ew_res;

	Vect_append_point(points, x, y);
    }

    return TRUE;
}

/****************************************************************************
* int CheckPos(ContourCell *current, int numRow, int numCol)
*
* PURPOSE: 	To
* INPUT VARS:	current
*   	    	numRow
*   	    	numCol
* RETURN VAL:	none
*****************************************************************************/
static int CheckPos(ContourCell *current, int numRow, int numCol)
{
    if ((current->row >= 0) && (current->row <= numRow - 2) &&
    	(current->col >= 0) && (current->col <= numCol - 2))
    {
	return TRUE;
    }

    return FALSE;
}

/****************************************************************************
* void ResetCell(ContourCell *current, DCELL **z)
*
* PURPOSE: 	To
* INPUT VARS:	current
*   	    	z
* RETURN VAL:	none
*****************************************************************************/
static void ResetCell(ContourCell *current, DCELL **z)
{
    current->z[0] =  z[current->row][current->col];
    current->z[1] =  z[current->row][current->col + 1];
    current->z[2] =  z[current->row + 1][current->col + 1];
    current->z[3] =  z[current->row + 1][current->col];

    return;
}

/****************************************************************************
* void NewEdge (ContourCell *current)
*
* PURPOSE: 	updates edge number and row and col to those of next cell
* INPUT VARS:	current
* RETURN VAL:	none
*****************************************************************************/
static void NewEdge (ContourCell *current)
{
    switch (current->edge)
    {
	case TOP_EDGE:
	    current->row -= 1;
	    current->edge = BOTTOM_EDGE;
	    break;
	
	case RIGHT_EDGE:
	    current->col += 1;
	    current->edge = LEFT_EDGE;
	    break;
	
	case BOTTOM_EDGE:
	    current->row += 1;
	    current->edge = TOP_EDGE;
	    break;
	
	case LEFT_EDGE:
	    current->col -= 1;
	    current->edge = RIGHT_EDGE;
	    break;
	
	default:
	    G_fatal_error ("Illegal edge number while finding new edge");
    }
fprintf(stderr, "End NewEdge new edge = %d row = %d col = %d\n", current->edge, current->row, current->col);
}

/****************************************************************************
* void SetEdge (ContourCell *current, int edge)
*
* PURPOSE: 	sets the edge value in current
* INPUT VARS:	current
* RETURN VAL:	none
*****************************************************************************/
static void SetEdge (ContourCell *current, int edge)
{
    switch (edge)
    {
	case TOP_EDGE:
	    current->edge = TOP_EDGE;
	    break;
	
	case RIGHT_EDGE:
	    current->edge = RIGHT_EDGE;
	    break;
	
	case BOTTOM_EDGE:
	    current->edge = BOTTOM_EDGE;
	    break;
	
	case LEFT_EDGE:
	    current->edge = LEFT_EDGE;
	    break;
	
	default:
	    G_fatal_error ("Illegal edge number while finding new edge");
    }
}

/****************************************************************************
* int FindCrossing (ContourCell *current, double level,
*   struct Cell_head cellHead, struct line_pnts *points, int noerr)
*
* PURPOSE: 	decides which edge exit point from cell is on and changes
*   	    	value of edge. Returns 1 if only 2 crossings found ( don't
*   	    	want to check this Cell again); 0 otherwise.
* INPUT VARS:	current
*   	    	level
*   	    	cellHead
*   	    	points
*   	    	noerr
* RETURN VAL:	cellHit
*****************************************************************************/
static int FindCrossing (ContourCell *current, double level,
    struct Cell_head cellHead, struct line_pnts *points, int noerr)
{
    double  mid;
    int     numCross;   	/* number of crossings found in this Cell */
    int     edgeHit[4]; 	/* hit flag for each edge of Cell */
    int     cellHit = FALSE;
    int     i, j;

    numCross = 0;
    
    for (i = 0; i < 4; i++)
    {
	edgeHit[i]=FALSE;
	edgeHit[i] = CheckEdge(current->z[i], current->z[(i + 1) % 4], level);
	
	if (edgeHit[i])
	{
	    numCross++;
	}
    }

fprintf(stderr, "numCross = %d\n", numCross);
    if (numCross == 2)
    {
	cellHit = TRUE;
	edgeHit[current->edge] = FALSE;
	
	for (j = 0; j < 4; j++)
	{
	    if (edgeHit[j])
	    {
		SetEdge(current, j);
		AddPoint(FALSE, current, level, cellHead, points);
		break;
	    }
	}
    }
    else if (numCross == 4)
    {
fprintf(stderr, "%10.8g, %10.8g, %10.8g, %10.8g\n", current->z[0], current->z[1], current->z[2], current->z[3]);
fprintf(stderr, "row = %d col = %d\n", current->row, current->col);
	if (current->edge == TOP_EDGE)
	{
	    cellHit = TRUE;
	}

	mid = (current->z[0] + current->z[1] + current->z[2] +
	    current->z[3]) / 4;
	
	if (CheckEdge(mid, current->z[current->edge], level))
	{
	    SetEdge(current, 
	    	(current->edge == TOP_EDGE) ? LEFT_EDGE : current->edge - 1);
	}
	else
	{
	    SetEdge(current,
	    	(current->edge == LEFT_EDGE) ? TOP_EDGE : current->edge + 1);
	}
	
	AddPoint(FALSE, current, level, cellHead, points);

	if (current->edge == TOP_EDGE)
	{
	    cellHit = TRUE;
	}
    }
    else 
    {
	if(!(noerr && (numCross == 1)))
	{
	    fprintf(stderr, "Error, %d crossings in Cell %d, %d\n", numCross,
	    	current->row, current->col);
fprintf(stderr, "%10.8g, %10.8g, %10.8g, %10.8g\n", current->z[0], current->z[1], current->z[2], current->z[3]);
	}
	
	cellHit = TRUE;
    }
    
fprintf(stderr, "End find crossing new edge = %d\n", current->edge);
    return cellHit;
}

/****************************************************************************
* void AddPoint (int first, ContourCell *curr, double level,
*   struct Cell_head cellHead, struct line_pnts *points)
*
* PURPOSE:  	finds crossing point using linear interpolation, converts
*   	    	from row-column to x-y space, and adds point to current line
* INPUT VARS:	first
*   	    	curr
*   	    	level
*   	    	cellHead
*   	    	points
* RETURN VAL:	none
*****************************************************************************/
static void AddPoint (int first, ContourCell *curr, double level,
    struct Cell_head cellHead, struct line_pnts *points)
{
    static double   prevX = -10.0;   /* -1.0 may be valid - one step off map */
    static double   prevY = -10.0;
    static double   startX = -10.0;
    static double   startY = -10.0;
    static int	    edge = -1;
    double  	    x, y;
    double  	    mapX;
    double  	    mapY;
    double  	    ratio;
    int     	    p1, p2;	

    p1 = curr->edge;
    p2 = (curr->edge + 1) % 4;
    
    if (G_raster_cmp(&curr->z[p1], &curr->z[p2], DCELL_TYPE) == 0)
    {
	ratio = 1.0;
    }
    else if (G_is_d_null_value(&curr->z[p1]))
    {
        ratio = 1.0 / 2.0;
    }
    else if (G_is_d_null_value(&curr->z[p2]))
    {
        ratio = 1.0 / 2.0;
    }
    else
    {
	ratio = (level - curr->z[p1]) / (curr->z[p2] - curr->z[p1]);
    }

/*fprintf(stderr,"edge = %d\n", curr->edge);*/
    switch (curr->edge) 
    {
	case TOP_EDGE:
	    y = curr->row;
	    x = curr->col + ratio;
	    break;

	case RIGHT_EDGE:
	    y = curr->row + ratio;
	    x = curr->col + 1;
	    break;

	case BOTTOM_EDGE:
	    y = curr->row + 1;
	    x = curr->col + 1 - ratio;
	    break;

	case LEFT_EDGE:
	    y = curr->row + 1 - ratio;
	    x = curr->col;
	    break;

	default:
	    G_fatal_error("Edge number out of range");
    }

    /* Convert r/c values to x/y values */
    if (x != prevX || y != prevY)
    {
fprintf(stderr, "AddPoint: new point row = %d, col = %d, x = %g, y = %g\n", curr->row, curr->col, x, y);
	prevX = x;
	prevY = y;
	
	if (!(HaveHit(x, y, CONTOUR_POINT)) || (x == startX && y == startY))
	{
	    AddHashNode(x, y, CONTOUR_POINT);
	
    	    mapY = cellHead.north - (y + 0.5) * cellHead.ns_res;
    	    mapX = cellHead.west + (x + 0.5) * cellHead.ew_res;

    	    Vect_append_point(points, mapX, mapY);
fprintf(stderr,"Point added\n");
	}
else
{
fprintf(stderr,"Point not added\n");
}

	if (first)
	{
fprintf(stderr, "start edge = %d, x = %g y = %g\n", curr->edge, x, y);
	    startX = x;
	    startY = y;
	}
    }
fprintf(stderr,"done AddPoint\n");
    
    return;
}

/****************************************************************************
* int CheckEdge (DCELL  d1, DCELL d2, double level)
*
* PURPOSE: 	returns TRUE if level is between values d1 & d2; FALSE
*   	    	otherwise.
* INPUT VARS:	d1
*   	    	d2
*   	    	level
* RETURN VAL:	boolean
*****************************************************************************/
static int CheckEdge (DCELL  d1, DCELL d2, double level)
{
    if (((d1 < level) && (d2 >= level)) || ((d1 >= level)  && (d2 < level)))
    {
	return TRUE;
    }

    return FALSE;
}

/*=========================== External Functions ===========================*/
			     
/****************************************************************************
* void contour (double levels[], int nlevels, struct Map_info Map, DCELL  **z,
*   struct Cell_head cellHead, FILE *Att, int quiet, int noerr, int numCut)
*
* PURPOSE: 	To generate contour...
* INPUT VARS:	evels[]
*   	    	nlevels
*   	    	Map
*   	    	z
*   	    	cellHead
*   	    	Att
*   	    	quiet
*   	    	noerr
*   	    	numCut
* RETURN VAL:	none
*****************************************************************************/
void contour (double levels[], int nlevels, struct Map_info Map, DCELL  **z,
    struct Cell_head cellHead, FILE *Att, int quiet, int noerr, int numCut)
{
    struct line_pnts	*points;
    ContourCell     	current;
    double  	    	level;	    /* current contour level */
    double  	    	x;  	    /* x value for line center for labels */
    double  	    	y;  	    /* y value for line center for labels */
    int     	    	numRow;	    /* number of rows in current region */
    int     	    	numCol;	    /* number of columns in current region */
    int     	    	startRow;   /* start row of current line */
    int     	    	startCol;   /* start col of current line */
    int     	    	outside;    /* flag to indicate contour outside map */
    int     	    	p1; 	    /* indexes to end points of cell edges */
    int     	    	p2; 	    /* indexes to end points of cell edges */
    int     	    	first;	    /* flag to indicate start of contour */
    int     	    	n, i, j;    /* loop counters */

    /* Allocate the hit hash table */
    AllocateHashTable();
    AllocateNodePool();
    
    points = Vect_new_line_struct();
    points->n_points = 0;

    numRow = cellHead.rows;
    numCol = cellHead.cols;

    if (!quiet)
    {
	fprintf(stderr, "Total levels: %4d   Current level: ", nlevels);
    }
    
    for (n = 0; n < nlevels; n++)
    {
	level = levels[n];
	
	if (!quiet)
    	{
	     /* Print level number */
	     fprintf(stderr, "%4d\b\b\b\b", n + 1);
	}

    	/* Initialize the hit hash table */
	InitHashTable();
	
	/* Check each cell of top and bottom borders  */
	for (startRow = 0; startRow < numRow; startRow += (numRow - 2))
/*	for (startRow = numRow - 2; startRow > 0; startRow = 0)*/
	{
	    for (startCol = 0; startCol <= numCol - 2; startCol++)
	    {

		/* Look for starting point of new line */
		if (!(HaveHit((double) startCol, (double) startRow,
		    CONTOUR_WINDOW)))
		{ 
		    current.row = startRow;
		    current.col = startCol;
		    outside = GetNewCell(&current, numRow, numCol, z, level,
		    	cellHead, points, startRow, startCol);

		    /* Is this top or bottom? */
		    if (startRow == 0)
		    {
			current.edge = TOP_EDGE;
		    }
		    else
		    {
			current.edge = BOTTOM_EDGE;
		    }
		    
		    p1 = current.edge;
		    p2 = current.edge + 1;

		    if (CheckEdge(current.z[p1], current.z[p2], level))
		    {
			first = TRUE;
			AddPoint(first, &current, level, cellHead, points);
			first = FALSE;
			
			/* While not off an edge, follow line */
			while (!outside)
			{
/*fprintf(stderr, "top-bot row = %d col = %d\n", current.row, current.col);*/
			    if (FindCrossing(&current, level, cellHead, points,
				noerr))
			    {
			    	AddHashNode((double) current.col,
				    (double) current.row, CONTOUR_WINDOW);
			    }
			    
			    NewEdge(&current);
			    outside = GetNewCell(&current, numRow, numCol, z,
			    	level, cellHead, points, startRow, startCol);
			}
/*fprintf(stderr, "line done\n");*/
			
			/* Skip the line if we don't have enough points */
			if ((numCut <= 0) || ((points->n_points) > numCut))
			{
			    Vect_write_line(&Map, LINE, points);
for (i = 0; i < points->n_points; i++)
{
    /*fprintf(stderr, "Point %d => x = %8.8g, y = %8.8g\n", i, points->x[i], points->y[i]);*/
}
			    get_line_center(&x, &y, points);
			    write_att(Att, 'L', x, y, (int)level);
			}
			
			Vect_destroy_line_struct(points);
                        points = Vect_new_line_struct();
		    } /* if checkedge */
		} /* if ! hit */
	    } /* for columns */
	} /* for rows */

	/* Check right and left borders (each row of first and last column) */
	for (startCol  = 0; startCol < numCol; startCol += (numCol - 2))
	{
	    for (startRow = 0; startRow <= numRow - 2; startRow++)
	    {
		/* Look for starting point of new line */
		if (!(HaveHit((double) startCol, (double) startRow,
		    CONTOUR_WINDOW)))
		{
		    current.row = startRow;
		    current.col = startCol;
		    outside = GetNewCell(&current, numRow, numCol, z, level,
		    	cellHead, points, startRow, startCol);

		    /* Is this left or right edge? */
		    if (startCol == 0)
		    {
			current.edge = LEFT_EDGE;
		    }
		    else
		    {
			current.edge = RIGHT_EDGE;
		    }
		    
		    p1 = current.edge;
		    p2 = (current.edge + 1) % 4;
		    
		    if (CheckEdge(current.z[p1], current.z[p2], level))
		    {
			first = TRUE;
			AddPoint(first, &current, level, cellHead, points);
			first = FALSE;
			
			/* While not off an edge, follow line */
			while (!outside)
			{
/*fprintf(stderr, "left-right row = %d col = %d\n", current.row, current.col);*/
			    if (FindCrossing(&current, level, cellHead, points,
				noerr))
			    {
			    	AddHashNode((double) current.col,
				    (double) current.row, CONTOUR_WINDOW);
			    }
			    
			    NewEdge(&current);
			    outside = GetNewCell(&current, numRow, numCol, z,
			    	level, cellHead, points, startRow, startCol);
			}
			
/*fprintf(stderr, "line done\n");*/
			/* Skip the line if we don't have enough points */
			if ((numCut <= 0) || ((points->n_points) > numCut))
			{
			    Vect_write_line(&Map, LINE, points);
for (i = 0; i < points->n_points; i++)
{
    /*fprintf(stderr, "Point %d => x = %8.8g, y = %8.8g\n", i, points->x[i], points->y[i]);*/
}
			    get_line_center(&x, &y, points);
			    write_att(Att, 'L', x, y, (int)level);
			}
			
			Vect_destroy_line_struct(points);
                        points = Vect_new_line_struct();
		    } /* if checkedge */
		} /* if ! hit */
	    } /* for rows */
	} /* for columns */

	/* Check each interior Cell */
	for (startRow = 1; startRow <= numRow - 3; startRow++)
	{ 
	    for (startCol = 1; startCol <= numCol - 3; startCol++)
	    {
		/* Look for starting point of new line */
		if (!(HaveHit((double) startCol, (double) startRow,
		    CONTOUR_WINDOW)))
		{
		    current.row = startRow;
		    current.col = startCol;
		    current.edge = TOP_EDGE;
		    outside = GetNewCell(&current, numRow, numCol, z, level,
		    	cellHead, points, startRow, startCol);
		    
		    if (CheckEdge(current.z[0], current.z[1], level))
		    {
/*fprintf(stderr, "heights = %g and %g row = %d col = %d\n", current.z[0], current.z[1], current.row, current.col);*/
			first = TRUE;
			AddPoint(first, &current, level, cellHead, points);
			first = FALSE;
			
			if (FindCrossing(&current, level, cellHead, points,
			    noerr))
			{
			    AddHashNode((double) current.col,
				(double) current.row, CONTOUR_WINDOW);
			}

			NewEdge(&current); 
			outside = GetNewCell(&current, numRow, numCol, z,
			    level, cellHead, points, startRow, startCol);

			/* While not back to starting point, follow line */
			while ((current.edge != TOP_EDGE) || 
			    ((current.row != startRow) ||
			    (current.col != startCol)))
			{
/*fprintf(stderr, "center row = %d col = %d edge = %d\n", current.row, current.col, current.edge);*/
			    if (FindCrossing(&current, level, cellHead, points,
				noerr))
			    {
			    	AddHashNode((double) current.col,
				    (double) current.row, CONTOUR_WINDOW);
			    }
			    
			    NewEdge(&current);
			    outside = GetNewCell(&current, numRow, numCol, z,
			    	level, cellHead, points, startRow, startCol);
			}
/*fprintf(stderr, "line done\n"); */
/*fprintf(stderr, "center row = %d col = %d edge = %d\n", current.row, current.col, current.edge);*/
			
			/* Skip the line if there was only 1 point */
			/* Skip the line if we don't have enough points */
			if ((numCut <= 0) || ((points->n_points) > numCut))
			{
			    Vect_write_line(&Map, LINE, points);
for (i = 0; i < points->n_points; i++)
{
    /*fprintf(stderr, "Point %d => x = %8.8g, y = %8.8g\n", i, points->x[i], points->y[i]);*/
}
			    get_line_center(&x, &y, points);
			    write_att(Att, 'L', x, y, (int)level);
			}
			
			Vect_destroy_line_struct(points);
                        points = Vect_new_line_struct();
		    } /* if checkedge */
		} /* if ! hit */
	    } /* for rows */
	} /* for columns */
    } /* for levels */

    fprintf(stdout, "       \n");
    Vect_destroy_line_struct(points);
}

